/*
 *  source code metrics counter
 *  Copyright (C) 2015-2016  NEC Corporation.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package com.nec.ncmeter;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import japa.parser.JavaParser;
import japa.parser.ParseException;
import japa.parser.TokenMgrError;
import japa.parser.ast.CompilationUnit;

import com.sun.jna.Library;
import com.sun.jna.Native;
import com.sun.jna.*;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;


public final class NCMJParser {

    static String target_filepath;                         /** Path of input java file. */        

    static String error_filepath;                          /** Path of 'temp_error_output.tmp' file. */
    
    static String error_filename;                          /** Path of error output. */ 

    static String metric_filepath;                         /** Path of file containing output. */
    
    final static String metric_filename = "javaOutput.csv"; /** Name of file containing output. */
    
    final static String xml_filename    = "javaData.xml";   /** Name of file containing xml output from 'ncmeter.exe'. */
    
    final static String file_separator = System.getProperty("file.separator");  /** File separator. */
    
    static String parse_filepath;                           /** Path of file containing output from nkf functionality. */	
    
    static String nkf_propValue = null;                     /** Property value denoting whether to implement nkf functionality. */
    
    static String nkf_exeOrDll = null;                 /** Flag that denotes whether to use nkf exe or nkf dll. */
        
    final static String parse_filename  = "tempnkf";        /** Name of file containing output from nkf functionality. */
    
    static String charset_name;                             /** Charset value. */                

    static int debug_level = 0;                             /** Field indicating whether execution information needs to be logged. */

    static int java_exit = 0;                               /** Exit value. */

    public static String crlf;                              /** Line separator. */
    
    static String xmlFilePath;                              /** Path of file containing the xml ouput from 'ncmeter.exe'. */
    
    static String output_dir;                               /** Location of the temp directory created by 'ncmeter.exe'. */

    static String func_opt;                                 /** Set to param. */

    static String comp_opt;                                 /** Version of java to be used. */
    
    static StringBuffer errorbuff ;                         /** Buffer to hold error information. */
    
    static ResourceBundle myResources;                      /** Data taken from 'CDMtrMessage.properties' file. */

    private static NCMStructData structdata;                /** Data type created from the compilation unit instance of an input file. */

    private static NCMNodeParser nodeParse;                 /** Extracts data from compilation unit instance into structdata instance. */
    
    static String fileTagName = "file";
    
    static String langOptTagName = "langOpt";
    
    static String charsetTagName = "charset";
    
    static String dirPathTagName = "dirPath";
    
    static String fileNameTagName = "fileName";
    
    static String filePathTagName = "inputFilePath";
    
    static BufferedWriter bw = null;
  
    static String nkfPath = null;                          /** Path of the nkf dll or exe, depending upon whether dll or exe needs to be used. */
    
    static int inum = 0;                                   /** Number of java input files successfully used. */
	
    static String long_path_prefix = "\\\\?\\";            /* This variable hold the file convention for long Path prefix */

    static String unc_long_path_prefix = "\\?\\UNC\\";     /* This variable hold the file convention for UNC long Path prefix */

    static int long_path_prefix_size = 4;                  /* This variable holds the length of prefix in the path */

    static int unc_long_path_prefix_size = 7;              /* This variable holds the length of UNC prefix in the path */

/** 
 *Interface for loading nkf.dll. Methods to be used are also declared here.
 */
    public interface NKF extends Library {
    	
	    NKF INSTANCE =(NKF) Native.loadLibrary("nkf32", NKF.class);                        /** 'nkf32' is the name of the dll file. */	
        int NkfFileConvert2SafeW(WString filea, int lengtha,WString fileb,int lengthb);
        void SetNkfOption(String option);      
    }
    
    /**
     * メイン処理
     */
    public static void main(String[] args) throws Exception {
        
    	crlf = System.getProperty("line.separator");
        errorbuff = new StringBuffer();
        try {
            myResources = ResourceBundle
                    .getBundle("com/nec/ncmeter/CDMtrMessage");
        } catch (Exception ex) {
            myResources = null;
        }
        
        /**
         * Parse command line arguments. 
         * This is done using the 'gnu getopt' library.
         * Various variables are allocated values.
         */
        if (!parseOptions(args, errorbuff)) {
            try {
                if (myResources != null) {
                    String msg00101 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000001_01");

                    outputLogFile(1, msg00101 + errorbuff.toString() + crlf); // MSG_LOG_ID_JAVA_00000001
                }
            } catch (Exception ex) {
                outputLogFile(1, ex + crlf);
            }
            errorbuff = null;
            System.exit(1);
        }
                
        /**
         * Parse xml file using SAX parser and make calls for processing each file
         * in the xml file.
         */
        try {

        	SAXParserFactory factory = SAXParserFactory.newInstance();
        	SAXParser saxParser = factory.newSAXParser();

        	
        	DefaultHandler handler = new DefaultHandler() {
            
        	/**
        	 * Variables representing the various tags that are present in the xml file produced by 'ncmeter.exe'.
        	 * Values of these represent whether the given tag is being processed(true) or not(false).
        	 */
        	boolean fileTag = false;
        	boolean langOptTag = false;
        	boolean charsetTag = false;
        	boolean dirPathTag = false;
        	boolean fileNameTag = false;
        	boolean filePathTag = false;
        	/** Flag to check if data for a single file was read correctly from xml file. */
        	boolean fileTagFlag = true;
        	/** String to hold the absolute path of the input file. */
        	StringBuffer fileAbsPath = new StringBuffer();
        	
        	/**
        	 * This method is called whenever the start of a tag is encountered.The boolean value for the corresponding
        	 * variable is set to 'true'. 
        	 */

        	public void startElement(String uri, String localName,String qName, 
                        Attributes attributes) throws SAXException {

        		if (qName.equalsIgnoreCase(fileTagName)) {                           /** 'file' tag encountered. */
        			fileTag = true;
         		}
        		
        		if (qName.equalsIgnoreCase(langOptTagName)) {                        /** 'langOpt' tag encountered. */
        			langOptTag = true;
        		}

        		if (qName.equalsIgnoreCase(charsetTagName)) {                        /** 'charset' tag encountered. */
        			charsetTag = true;
        		}

        		if (qName.equalsIgnoreCase(dirPathTagName)) {                        /** 'dirPath' tag encountered. */
        			dirPathTag = true;
        		}

        		if (qName.equalsIgnoreCase(fileNameTagName)) {                      /** 'fileName' tag encountered. */
        			fileNameTag = true;
        		}
        		        		
        		if (qName.equalsIgnoreCase(filePathTagName)) {                       /** 'inputFilePath' tag encountered. */
        			filePathTag = true;
        		}

        	}

        	/**
        	 * This method is called whenever an ending tag is encountered. If the ending tag is 'file', then the file 
        	 * corresponding to this tag will be processed by the call to 'calculateMetrics()'.
        	 */
        	public void endElement(String uri, String localName,
        		String qName) throws SAXException {
                 
        		if (qName.equalsIgnoreCase(fileTagName) && fileTag) {
        			if(!(new File(fileAbsPath.toString()).exists())){	
        				errorbuff.append("Input file doesn't exists : " + target_filepath);
        				fileTagFlag=false;
        			}
        			if(fileTagFlag){
        			/**Calculate structure metrics for the given file */
	        			try {
	        				target_filepath=fileAbsPath.toString();
	        				fileAbsPath.delete(0, fileAbsPath.length());
							calculateMetrics();
						} catch (IOException ioe) {
							errorbuff.append(ioe.getMessage() + crlf);
				            outputLogFile(1, errorbuff.toString()); 
						} catch(InterruptedException ie){
							errorbuff.append(ie.getMessage() + crlf);
				            outputLogFile(1, errorbuff.toString());
						} catch(Error e){
							errorbuff.append(e.getMessage() + crlf);
				            outputLogFile(1, errorbuff.toString());
							System.gc();
						}
        			}
        			
        			fileTag = false;
        		}else if(qName.equalsIgnoreCase(langOptTagName) && langOptTag) {               /** Do nothing for this tag is not required in java. */
        			
        		    langOptTag = false;
        		}else if(qName.equalsIgnoreCase(charsetTagName) && charsetTag) {               /** No use of this tag in the java parser. */
        			
        			charsetTag = false;
        		}else if(qName.equalsIgnoreCase(dirPathTagName) && dirPathTag) {               /** No use of this tag in the java parser. */         			
        			
        			dirPathTag = false;
        		}else if(qName.equalsIgnoreCase(fileNameTagName) && fileNameTag) {             /** No use of this tag in the java parser. */
        			
        			fileNameTag = false;
        		}else if(qName.equalsIgnoreCase(filePathTagName) && filePathTag) {             /** No use of this tag in the java parser. */
        			
        			filePathTag = false;
        		}

        	}

        	/**
        	 * This method extracts the value inside the tag that was last encountered and whose boolean variable was set to 'true'.
        	 */
        	public void characters(char ch[], int start, int length) throws SAXException {        		
        		
                if (charsetTag) {               /** Assign variable concerned with charset value. */
        			
        			charset_name = (new String(ch, start, length)).toString();
        			
        		}else if (filePathTag) {              /** Store full path of the input file into target_filepath. */
        			
        			fileAbsPath.append((new String(ch, start, length)).toString());
        			        			     			
        		}
        	}
           };
               
             /**
              * Parse the file pointed to by 'xmlFilePath' using the handler defined above.
              */
              saxParser.parse(xmlFilePath, handler);                                    
              StringBuffer csvbuff = new StringBuffer();
              FileOutputStream fo = null;
              OutputStreamWriter fw = null;
              fo = new FileOutputStream(metric_filepath, true);
              fw = new OutputStreamWriter(fo, "UTF-8");
              csvbuff.append("\"Value of iNumOfSuccess\",\"random\",-,JAVA,\"" + inum + ",-,-,-,-,-,-,-,-,-,-,-,-" + crlf);
            
              fw.write(csvbuff.toString());
              fw.close();
             } catch (Exception ex) {
            	 errorbuff.append(ex.getMessage() + crlf);
		         outputLogFile(1, errorbuff.toString());
             }
       System.exit(java_exit);
    }

/** This method calculates the structure metrics of the input file. First nkf functionalities are performed
  * on the input file and 'parse_filepath' is populated with the output.Then an instance of compilation 
  * unit is created from the nkf ouput. Further, from this compilation unit, an instance of 
  * 'NCMStructData'(structData) is created and finally the information from structData is populated
  *  in the output file(javaOutput.csv). 
  * @throws InterruptedException 
  */
    
    private static void calculateMetrics() throws IOException, InterruptedException{
    	
    	 
        CompilationUnit cu = null;
        FileInputStream in = null;
        int iRet;

        try {
        /** Store the path of input file with long path prefix. */ 
        String nkfInputFile = target_filepath; 
        /** Remove the long path prefix from the input file path. */ 
        target_filepath = removeLongPath(target_filepath); 
        
        /** Check if nkf functionality needs to be implemented. */
        if(nkf_propValue.equalsIgnoreCase("true")){
            /**Perform nkf functionality and produce output file(parse_filepath) in the 'temp_dir'. */
            iRet = performNKF(nkfInputFile);
        }
        else{
        	/** Process the input file instead. */
            parse_filepath = target_filepath;
        }
                        
        /**Process the 'tempnkf' file */
      

            in = new FileInputStream(parse_filepath);

            /**Obtain the compilation unit instance. */
            if (charset_name == null)
                cu = JavaParser.parse(in);
            else
                cu = JavaParser.parse(in, charset_name);
                	
        } catch (FileNotFoundException ex) {

            errorbuff.append(ex.getMessage() + crlf);
            outputLogFile(1, errorbuff.toString()); // MSG_LOG_ID_JAVA_00000002
            cu = null;
        } catch (ParseException ex) {
            try {	
                if (myResources != null) {
                    String msg00301 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_01");
                    String msg00302 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_02");
                    String msg00303 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_03");
                    String msg00304 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_04");
                    String msg00305 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_05");
                    String msg00306 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_06");
                    String msg00307 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_07");
                    String msg00308 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_08");
                    String msg00309 = myResources
                            .getString("MSG_LOG_ID_JAVA_00000003_09");

                    errorbuff.append(msg00301 + crlf + msg00302
                            + ex.currentToken.beginLine + msg00303
                            + ex.currentToken.beginColumn);
                    errorbuff.append(msg00304 + ex.currentToken.image
                            + msg00305 + crlf);
                    if (ex.currentToken.next != null) {
                        errorbuff.append(msg00306
                                + ex.currentToken.next.beginLine + msg00307
                                + ex.currentToken.next.beginColumn);
                        errorbuff.append(msg00308 + ex.currentToken.next.image
                                + msg00309 + crlf);
                    }
                    errorbuff.append(ex.getMessage() + crlf);
                    outputLogFile(1, errorbuff.toString()); // MSG_LOG_ID_JAVA_00000003
                }
            } catch (Exception ex2) {
                outputLogFile(1, ex2 + crlf);
            }
            cu = null;
        } catch (TokenMgrError ex) {

            errorbuff.append(ex.getMessage() + crlf);
            outputLogFile(1, errorbuff.toString()); // MSG_LOG_ID_JAVA_00000004
            cu = null;
        } finally {

            try {
                if (in != null)
                    in.close();
                    in = null;
            } catch (Exception e) {
                errorbuff.append(e.getMessage() + crlf);
                outputLogFile(1, errorbuff.toString()); // MSG_LOG_ID_JAVA_00000005
            }
        }
 
        /** Create an instance of NCMStructData from the compilation unit instance just obtained. */
        structdata = new NCMStructData();
        structdata.DirPath = getPathName(target_filepath);
        structdata.FileName = getFileName(target_filepath);
        structdata.LangType = "JAVA";
        structdata.debugLevel = debug_level;
        structdata.funcOpt = func_opt;

        if (cu != null) {

            nodeParse = new NCMNodeParser();
            try {
            	
                /** Parse the compilation unit instance using the nodeParse instance and populate the structdata instance. */
                nodeParse.visit(cu, structdata);
            } catch (StackOverflowError ex) {
                try {
                    if (myResources != null) {
                        String msg00601 = myResources
                                .getString("MSG_LOG_ID_JAVA_00000006_01");

                        errorbuff.append(msg00601 + crlf);
                        outputLogFile(1, errorbuff.toString()); // MSG_LOG_ID_JAVA_00000006

                    }
                } catch (Exception ex3) {
                    outputLogFile(1, ex3 + crlf);
                }
                structdata.errorLevel = 2;
                java_exit = 1;
            }
            /** Output the execution information into the 'temp_err_output.tmp' file. */
            outputLogFile(2, structdata.dbugbufread()); // MSG_LOG_ID_JAVA_00000007
        } else {
            structdata.errorLevel = 1;
            java_exit = 1;
        }

        /** Output 'structdata' information into the output file. */
        if (!outputMetricFile(structdata, errorbuff)) {

            errorbuff.append(errorbuff);
            outputLogFile(1, errorbuff.toString()); // MSG_LOG_ID_JAVA_00000008
            java_exit = 1;
        }  	
    }
    	
    
    private static int performNKF(String nkfInputFile) throws IOException, InterruptedException {
		
        /** Decide whether to use nkf exe or dll. */       
        if(nkf_exeOrDll.equals("EXE")){
            /** Execute the nkf.exe on 'target_filepath'. */
            performNKFExe(nkfInputFile);
        }else if(nkf_exeOrDll.equals("DLL")){
            /** Execute the nkf.dll on 'target_filepath'. */
            performNKFDll(nkfInputFile);
        }else{
           /** Do some error handling. */
           errorbuff.append("nkf failed");
    	}
        return 0;
    }

    private static int performNKFDll(String nkfInputFile) throws IOException {
		
        /** Value to return. */
    	int retVal=0;
        
    	try{ 
        
        /**  Removing redundant dots('.' , '..') from the file path. */
        nkfPath = (new File(nkfPath)).getCanonicalPath();
        
        /** Set the path to the nkf32.dll file */
    	System.setProperty("jna.library.path",nkfPath);
    	NKF sdll = NKF.INSTANCE;
    	
        Boolean ret=false;
       
        /** Marshalling target_filepath and parse_filepath. */
        WString fileA = new WString(nkfInputFile);
        WString fileB = new WString(parse_filepath);
        
        /** Call nkf methods. */
        
        /** SetNkfOption sets some sort of option. */
        sdll.SetNkfOption("-Lw -w80");
        
        /** NkfFileConvert2SafeW takes the given input file, processes it and produces the output in the 'tempnkf' file. */
        retVal =  sdll.NkfFileConvert2SafeW(fileA,fileA.length()+1,fileB,fileB.length()+1);       
        }catch(Exception e){
        	e.printStackTrace();
        }        
    	return retVal;
    }

 
    private static int performNKFExe(String nkfInputFile) throws IOException, InterruptedException{
	
    	/** List to contain the command line argument. */
    	List<String> cmd = new ArrayList<String>();
    	/** Instantiate the process builder object. */
    	ProcessBuilder process = new ProcessBuilder();
    	/** Prepare the command line argument. */
    	/** Split the path nkf using escape character as delimiter. */
    	String[] nkfSplit = nkfPath.split("\\\\");
    	/** Absolute path of the nkf.exe file. */
    	String nkfAbsPath = new String("");
    	/** Create absolute path by joining back strings previously split
    	 *  using the escape character.
    	 */
    	for(int i=0;i<nkfSplit.length-1;i++){
    		nkfAbsPath = nkfAbsPath.concat(nkfSplit[i]);
    		nkfAbsPath = nkfAbsPath.concat("\\");
    	}
    	String[] splitOption = nkfSplit[nkfSplit.length-1].split(" ", 2);
    	nkfAbsPath = nkfAbsPath.concat(splitOption[0]);
    	nkfAbsPath = "\"" + nkfAbsPath + "\"" ;
    	/** Absolute path created, now add to the command list. */
    	cmd.add(nkfAbsPath);
    	cmd.add(splitOption[1]);
    	/** Path of the input file to be processed. */
        cmd.add(nkfInputFile);    	
    	/** Attach the command list with the process builder instance. */
    	process.command(cmd);
    	/** Redirect the output to the nkf output file. */
    	process.redirectOutput(new File(parse_filepath));
    	/** Redirect the error output to the error output file. */
    	process.redirectError(new File(error_filepath));
    	/** Invoke nkf.exe. */
    	(process.start()).waitFor();
	    return 0;	
    }
    
	/**
     * ファイルのフルパス名からファイル名を取得
     */
    private static String getFileName(String filename) {

        int index = filename.lastIndexOf(File.separator);
        if (index > -1)
            return filename.substring(index + 1);
        else
            return filename;
    }

    /**
     * パースを行う際のJavaバージョン情報を取得
     */
    public static String getCompOpt() {
        return comp_opt;
    }

    /**
     * ファイルのフルパス名からパス名を取得
     */
    public static String getPathName(String filename) {
        int index = filename.lastIndexOf(File.separator);
        if (index > 0)
            return filename.substring(0, index);
        else
            return "";
    }

    /**
     * オプションを解析します。 引数を受け取りクラス変数にオプションに指定された情報を設定します。 パラメータ解析処理はgnu.getoptを利用。
     */
    private static boolean parseOptions(final String[] args,
            StringBuffer errorbuf) {
        boolean res = true;
        int c;
        
        /** Following section creates an instance of LongOpt that is used while parsing the command line arguments that have been passed by 'ncmeter.exe'. */
        LongOpt[] longopts = { new LongOpt("func_opt", LongOpt.REQUIRED_ARGUMENT, null,'f'),
        		               new LongOpt("nkf_prop", LongOpt.REQUIRED_ARGUMENT, null,'n'),
        		               new LongOpt("nkf_ExeOrDll", LongOpt.REQUIRED_ARGUMENT, null,'o'),
        		               new LongOpt("nkf_path", LongOpt.REQUIRED_ARGUMENT, null,'l'),
        		               new LongOpt("comp_opt", LongOpt.REQUIRED_ARGUMENT, null,'p'),
        		               new LongOpt("output_dir", LongOpt.REQUIRED_ARGUMENT, null,'d'),
        		               new LongOpt("debug", LongOpt.NO_ARGUMENT, null, 2),
        		               new LongOpt("debug2", LongOpt.NO_ARGUMENT, null, 3),
        		               new LongOpt("error_output", LongOpt.REQUIRED_ARGUMENT,null, 'e')};
                    
        /** Getopt instance helps getting the value of all the cmd arguments. */
        Getopt g = new Getopt("ncmeter", args, "-:e:f:p:d:n", longopts);
        g.setOpterr(false);

        /** 'getopt()' will return an integer value corresponding to the longopts field. For eg:- 'e' if '--error_output' is encountered. */
        while ((c = g.getopt()) != -1) {
            switch (c) {	
            
            case 2:
                if (debug_level == 0)
                    debug_level = 1;
                break;
            case 3:
                if (debug_level <= 1)
                    debug_level = 2;
                break;	
            case 'e':
                if (error_filename == null) {
                	error_filename = g.getOptarg();
                } else {
                    errorbuf.append("-" + (char) c);
                    res = false;
                }
                break;
            case 'n':
            	if (nkf_propValue == null) {
            		nkf_propValue = g.getOptarg();
                } else {
                    errorbuf.append("-" + (char) c);
                    res = false;
                }
            	break;
            case 'o':
            	if (nkf_exeOrDll == null) {
            		nkf_exeOrDll = g.getOptarg();
                } else {
                    errorbuf.append("-" + (char) c);
                    res = false;
                }
            	break;
            case 'l':
            	if (nkfPath == null) {
            		nkfPath = g.getOptarg();
                } else {
                    errorbuf.append("-" + (char) c);
                    res = false;
                }
            	break;
            case 'f':
                if (g.getOptarg().compareTo("param") == 0) {
                    func_opt = "param";
                } else if (g.getOptarg().compareTo("none") == 0) {
                    func_opt = "none";
                } else {
                    errorbuf.append("-" + (char) c);
                    res = false;
                }
                break;
            case 'p':
                if (comp_opt == null) {
                    comp_opt = g.getOptarg();
                } else {
                    errorbuf.append("-" + (char) c);
                    res = false;
                }
                break;                        	                
            case 'd':
            	 if(output_dir==null){
            		 output_dir=g.getOptarg();
            	 }else{
            		 errorbuf.append("-" + (char) c);
            		 res=false;
            	 }
                
            }
                if (!res)
                break;
        }

        longopts = null;
        g = null;
        if (res) {     
        	 if (func_opt == null) {
                func_opt = "param";
            } else {
            }
        }
        
        String dir_path = output_dir + file_separator;
        
       /** Set path of file having output in csv. */
        metric_filepath = dir_path + metric_filename;
        
        /** Set path of file having output from nkf. */
        parse_filepath = dir_path + parse_filename;
        
        /** Set path of file having error output. */
        error_filepath = dir_path + error_filename;
        
        /** Set path of file having xml input. */
        xmlFilePath = dir_path + xml_filename;
        return (res);
    }

    /**
     * 文字列をバイト配列に変換する。
     */
    public static byte[] StringToByteArray(String hex) {

        byte[] bytes = new byte[hex.length() / 2];

        for (int index = 0; index < bytes.length; index++) {

            try {
                bytes[index] = (byte) Integer.parseInt(
                        hex.substring(index * 2, (index + 1) * 2), 16);
            } catch (Exception ex) {
                outputLogFile(1, ex.getMessage()); // MSG_LOG_ID_JAVA_00000009
                bytes = null;
                break;
            }
        }
        return bytes;
    }

    /**
     * ファイルにログを出力
     */
    private static void outputLogFile(final int idebug, final String buf) {
        FileOutputStream fo = null;
        OutputStreamWriter fw = null;
        if (debug_level >= idebug) {
            try {
                fo = new FileOutputStream(error_filepath, true);
                fw = new OutputStreamWriter(fo);
                fw.write(buf);
            } catch (IOException ex) {
                System.out.println(ex); // MSG_LOG_ID_JAVA_00000010

            } finally {
                try {
                    if (fw != null) {
                        fw.close();
                        fw = null;
                    }
                } catch (IOException ex) {

                    System.out.println(ex); // MSG_LOG_ID_JAVA_00000011
                } finally {
                    try {
                        if (fo != null) {
                            fo.close();
                            fo = null;
                        }
                    } catch (IOException ex) {

                        System.out.println(ex); // MSG_LOG_ID_JAVA_00000012
                    }
                }
            }
        }
        return;
    }

    /**
     * メトリクス情報出力
     */
    private static boolean outputMetricFile(final NCMStructData structdata,
            StringBuffer errbuf) {
        FileOutputStream fo = null;
        OutputStreamWriter fw = null;
        NCMStructData.NCMSFuncData fncdata;
        StringBuffer csvbuff = new StringBuffer();
        boolean res = true;

        if (structdata.errorLevel > 0) {

        } else {
        	
                

            int functionCount = structdata.functionCount();            /** Get number of methods in the file. */
            if (functionCount == 0) {

                csvbuff.append("\"" + structdata.DirPath + "\",\""
                        + structdata.FileName + "\"");
                csvbuff.append(",-,JAVA");
                csvbuff.append(",-,-,-,-,-,-,-,-,-,-,-,-,-" + crlf);
            } else {
                for (int ii = 0; ii < functionCount; ii++) {
                    fncdata = structdata.functionGetData(ii);
                    csvbuff.append("\"" + structdata.DirPath + "\",\""
                            + structdata.FileName + "\",");
                    if (func_opt.compareTo("param") == 0) {

                        String str_func_name;
                        str_func_name = fncdata.nameFunc;
                        int length = fncdata.param_string.length();
                        if (length > 0) {
                            str_func_name = str_func_name
                                    + fncdata.param_string;
                        } else {
                            str_func_name = str_func_name + "(" + ")";
                        }
                        csvbuff.append("\"" + str_func_name + "\",JAVA");
                    } else {

                        csvbuff.append("\"" + fncdata.nameFunc + "\",JAVA");
                    }
                    csvbuff.append("," + fncdata.cntFloc + ","
                            + fncdata.cntEloc + "," + fncdata.cntCloc + ","
                            + fncdata.cntBloc);
                    csvbuff.append("," + fncdata.cntCyclo + ","
                            + fncdata.cntCycloWC + "," + fncdata.cntBranchs
                            + "," + fncdata.cntCndOps + "," + fncdata.cntNests
                            + ",-,-,-,-" + crlf);
                }
            }
        }
        fncdata = null;
        try {
            fo = new FileOutputStream(metric_filepath, true);
            fw = new OutputStreamWriter(fo, "UTF-8");

            fw.write(csvbuff.toString());
           if(csvbuff.length()>0)
        	   inum=inum+1;
        
        
        } catch (IOException ex) {
            errbuf.append(ex.getMessage());
            res = false;
        } finally {
        	
            csvbuff = null;
            try {
                if (fw != null) {
                    fw.close();
                    fw = null;
                }
            } catch (IOException ex) {
                errbuf.append(ex.getMessage());
                res = false;
            } finally {
                try {
                    if (fo != null) {
                        fo.close();
                        fo = null;
                    }
                } catch (IOException ex) {
                    errbuf.append(ex.getMessage());
                    res = false;
                }
            }
        }
   
        return (res);
      
    }

    /*
     * @param filename  Filename from which the longpath 
     *                  prefix needs to be removed. 
     *                   
     * @return          Filename with removed longpath 
     *                  prefix.  
     * @throws FileNotFoundException 
     */ 
     public static String removeLongPath(String filename) throws FileNotFoundException{ 
     if(filename.startsWith(long_path_prefix)) 
          return filename.substring(long_path_prefix_size);
     else if(filename.startsWith(unc_long_path_prefix)) 
          return filename.substring(unc_long_path_prefix_size);
     else
          throw new FileNotFoundException();
     } 
}
